mod api;
mod config;
mod model;
mod state;
mod storage_manager;

use crate::config::PicOxConfig;
use crate::model::album::Album;
use crate::model::image::ImageData;
use crate::state::Context;
use crate::storage_manager::StorageManager;
use axum::extract::{Multipart, Path, Query, State};
use axum::http::StatusCode;
use axum::response::IntoResponse;
use axum::routing::{get, post};
use axum::{Json, Router};
use j_db::database::Database;
use j_db::model::JdbModel;
use log::info;
use serde::{Deserialize, Serialize};
use std::path::PathBuf;
use std::sync::Arc;
use structopt::StructOpt;
use tokio::sync::RwLock;

type PicContext = Arc<Context>;

#[derive(Debug, Clone, StructOpt)]
struct Args {
    pub config: PathBuf,
}

#[derive(Debug, Clone, Serialize, Deserialize)]
struct CreateAlbum {
    pub album_name: String,
}

#[derive(Debug, Clone, Serialize, Deserialize)]
struct AlbumQuery {
    pub album_name: Option<String>,
}

#[derive(Debug, Clone, Serialize, Deserialize)]
struct AddImage {
    pub album: AlbumQuery,
    pub image_name: String,
}

async fn create_album(
    State(context): State<PicContext>,
    Json(album): Json<CreateAlbum>,
) -> impl IntoResponse {
    info!("Creating new album '{}'", album.album_name);

    let new_album = Album::new(&album.album_name, Vec::new(), 0);

    let new_album = context.db.insert(new_album).unwrap();

    (StatusCode::OK, Json(new_album))
}

async fn get_album(album_id: Path<u64>, State(context): State<PicContext>) -> impl IntoResponse {
    let album = context.db.get::<Album>(*album_id).unwrap();

    (StatusCode::OK, Json(album))
}

async fn add_image(
    State(context): State<PicContext>,
    mut img_data: Multipart,
) -> impl IntoResponse {
    let mut data: Vec<u8> = Vec::new();
    let mut metadata: Option<AddImage> = None;
    while let Some(field) = img_data.next_field().await.unwrap() {
        let field_name = field.name().clone();
        if let Some(field_name) = field_name {
            if field_name == "metadata" {
                let metadata_json = field.text().await.unwrap();
                metadata = Some(serde_json::from_str(&metadata_json).unwrap())
            } else if field_name == "img_data" {
                data.extend_from_slice(field.bytes().await.unwrap().as_ref());
            }
        }
    }

    let mut album =
        Album::find_album_by_query(&context.db, metadata.clone().unwrap().album).unwrap();

    let mut store_manager = context.store_manager.write().await;
    let img = store_manager
        .store_img(
            &context.db,
            None,
            ImageData::Bytes(data),
            &metadata.unwrap().image_name,
            0,
        )
        .await
        .unwrap();

    album.images.push(img.id().unwrap());

    context.db.insert::<Album>(album).unwrap();

    StatusCode::OK
}

async fn query_album(
    album_query: Query<AlbumQuery>,
    State(context): State<PicContext>,
) -> impl IntoResponse {
    let resp = Album::find_album_by_query(&context.db, album_query.0);
    (StatusCode::OK, Json(resp))
}

#[tokio::main]
async fn main() {
    let args = Args::from_args();

    let config = PicOxConfig::new(args.config);

    let db = Database::new(&config.db_path).unwrap();

    let store_manager = StorageManager::new(config.storage_config.clone());

    let context = Context {
        db,
        config: config.clone(),
        store_manager: RwLock::new(store_manager),
    };

    let context = Arc::new(context);

    // initialize tracing
    tracing_subscriber::fmt::init();
    // build our application with a route
    let app = Router::new()
        // `GET /` goes to `root`
        .route("/api/album/create", post(create_album))
        .route("/api/album/:id", get(get_album))
        .route("/api/album/", get(query_album))
        .route("/api/image/", post(add_image))
        .with_state(context);

    // run our app with hyper, listening globally on port 3000
    let listener = tokio::net::TcpListener::bind(&config.host).await.unwrap();
    info!("Serving at {}", config.host);
    axum::serve(listener, app).await.unwrap();
}
